/***** Includes ********/

#include <stdio.h>
#include <strings.h>
#include <math.h>

/***** Global declarations ******/

#define MAX_N  64
#define MAX_FACTS 1000
#define TAG_SIZE  2		       /* number of chars in tag field */
#define num_words n

#ifndef TRUE
#define TRUE  1
#endif

#ifndef FALSE
#define FALSE   0
#endif

static int n;			       /* number of neurons actually used
				        * during run */
int debugging = FALSE;
int resizing = FALSE;

struct a_fact {
  char pattern[MAX_N];
  char tag[TAG_SIZE];
} facts[MAX_FACTS];
static int next_fact = 0;

/*		Back Propagation globals		*/

#define prompt "BACK-PROP> "
#define MAX_NEURONS_PER_LAYER 64       /* better be same as MAX_N */
#define MAX_LAYER 4
#define MAX_LONG 2147483647	       /* 2^31-1 */
#define MAX_STRLEN 128
#define sign(x) (x<0 ? -1 : (x>0 ? 1 : 0))

typedef struct {		       /* extra connection is to hardwired
				        * "ON" unit */
  float connection[MAX_NEURONS_PER_LAYER + 1];
  float delta[MAX_NEURONS_PER_LAYER + 1];
  float output;
} a_site;
typedef a_site a_net[MAX_LAYER][MAX_NEURONS_PER_LAYER];
typedef char bool;
typedef struct {
  int length;
  float value[MAX_NEURONS_PER_LAYER];
} expanded_string;
typedef float float_vector[MAX_NEURONS_PER_LAYER + 1];

bool show_prompts;
int seed;
int nl[MAX_LAYER], nlayer;
float learning_rate, momentum_rate;
a_net net;

/************** Top Level driver loop ********************/

main()
{
  char line[MAX_STRLEN];	       /* character input line */
  int new;
  int which_fact;
  struct a_fact recall_fact;
  FILE *fp;
  char name[50], which_tag[TAG_SIZE];
  int i, j, k;
  int check_every, correct_count, sweep, sweep_max;
  bool correct, word_correct;
  float epsilon;
  int try, try_max, wc_count;

  printf("Pattern generation program\n\n");

  printf("Initializing system...");
  n = MAX_N;
  printf("n=%d...", n);
  printf("done\n\n");

  do {
    printf("%s ", prompt);
    if (fgets(line, MAX_STRLEN, stdin) == NULL)
      break;
    switch (line[0]) {
      case '\0':
	printf("\n");
      case '\n':
	continue;		       /* with do, not switch */
      case 'x':
	break;			       /* exit */
      case 'c':
	printf("Clearing storage...");
	next_fact = 0;
	initialize_network();
	printf("done!\n");
	continue;
      case 'A':		       /* store memory in archive */
	printf("Archiving memory...Enter name of file for archive\n");
	scanf("%s", name);
	fp = fopen(name, "w");
	printf("Archiving...");
	fwrite(&next_fact, sizeof (next_fact), 1, fp);
	fwrite(&facts[0], sizeof (struct a_fact), next_fact, fp);
	fclose(fp);
	printf("Done! archived %d facts\n", next_fact);
	continue;
      case 'B':		       /* retrieving memory from archive */
	printf("Retrieving memory from archive...Enter name of file\n");
	scanf("%s", name);
	fp = fopen(name, "r");
	printf("Retrieving...");
	fread(&next_fact, sizeof (next_fact), 1, fp);
	fread(&facts[0], sizeof (struct a_fact), next_fact, fp);
	fclose(fp);
	printf("Done! retrieved %d facts\n", next_fact);
	continue;
      case 'C':		       /* store network in archive */
	printf("Archiving network...Enter name of file for archive\n");
	scanf("%s", name);
	fp = fopen(name, "w");
	printf("Archiving...");
	fwrite(&nlayer, sizeof (nlayer), 1, fp);
	fwrite(&learning_rate, sizeof (learning_rate), 1, fp);
	fwrite(&momentum_rate, sizeof (momentum_rate), 1, fp);
	fwrite(&nlayer, sizeof (nlayer), 1, fp);
	fwrite(&nl[0], sizeof (nl), 1, fp);
	fwrite(net, sizeof (a_net), 1, fp);
	fclose(fp);
	printf("Network archived in %s!\n", name);
	continue;
      case 'D':		       /* retrieving network from archive */
	printf("Retrieving network from archive...Enter name of file\n");
	scanf("%s", name);
	fp = fopen(name, "r");
	printf("Retrieving...");
	fread(&nlayer, sizeof (nlayer), 1, fp);
	fread(&learning_rate, sizeof (learning_rate), 1, fp);
	fread(&momentum_rate, sizeof (momentum_rate), 1, fp);
	fread(&nlayer, sizeof (nlayer), 1, fp);
	fread(&nl[0], sizeof (nl), 1, fp);
	fread(net, sizeof (a_net), 1, fp);
	fclose(fp);
	printf("Retrieved network from %s!\n", name);
	continue;
      case 'E':
	printf("File to output to in \"back\" format?\n");
	scanf("%s", name);
	fp = fopen(name, "w");
	printf("Writing network to %s...\n", name);
	fprintf(fp, "%d		random number seed\n", seed);
	fprintf(fp, "%d		number of layers\n", nlayer);
	for (i = 0; i < nlayer; i++)
	  fprintf(fp, "%d		# in layer %d\n", nl[i], i);
	fprintf(fp, "%4.2f		learning rate\n", learning_rate);
	fprintf(fp, "%4.2f		momentum rate\n", momentum_rate);
	fprintf(fp, "%d		# of presentations\n", next_fact);
	for (i = 0; i < next_fact; i++) {
	  fprintf(fp, "m		presentation %d\n", i);
	  for (j = 0; j < nl[0]; j++) {
	    fprintf(fp, "%1d ", facts[i].pattern[j]);
	    if (j % 8 == 7)
	      fprintf(fp, "\n");
	  }
	  if (j % 8 != 0)
	    fprintf(fp, "\n");
	  fprintf(fp, "m\n");
	  for (j = 0; j < nl[nlayer - 1]; j++) {
	    if (facts[i].tag[0] - 'A' == j)
	      fprintf(fp, "1 ");
	    else
	      fprintf(fp, "0 ");
	    if (j % 26 == 25)
	      fprintf(fp, "\n");
	  }
	  if (j % 26 != 0)
	    fprintf(fp, "\n");
	}
	fclose(fp);
	printf("Network written in \"back\" format!\n");
	continue;
      case 'e':
	printf("Enter which fact you want erased\n");
	scanf("%d", &which_fact);
	erase_fact(which_fact);
	printf("Done!\n");
	continue;
      case 'f':
	printf("Currently have %d facts stored in memory\n", next_fact);
	continue;
      case 'd':
	if (debugging) {
	  printf("Debugging turned off\n");
	  debugging = FALSE;
	}
	else {
	  printf("Debugging turned on\n");
	  debugging = TRUE;
	}
	continue;
      case 'p':		       /* print network connections */
	printf("Network connections:\n");
	for (i = nlayer - 1; i > 0; i--) {
	  for (j = 0; j < nl[i]; j++) {
	    for (k = 0; k <= nl[i - 1]; k++) {
	      printf("%4.2f ", net[i][j].connection[k]);
	    }
	    printf("** ");
	  }
	  printf("\n");
	}
	continue;
      case 'g':		       /* get parameters */
	get_parameters();
	continue;
      default:
	help();
	continue;
    }
    break;
  } while (TRUE);

  printf("\nIt's been real!\n");
}

/****** Assorted subroutines *********************/

help()
{				       /* print out commands available */
  printf("Commands are:\n");
  printf("  s <tag>  Save fact with given <tag>\n");
  printf("  r <k>    Show recall of fact k\n");
  printf("  h <k>    Show hidden layer k outputs and connections\n");
  printf("  m <k>    Memorize facts, showing status every k sweeps\n");
  printf("  w <k>    Sweep through facts k times\n");
  printf("  p        Print network connections\n");
  printf("  g        Get net parameters\n");
  printf("  d        Toggle debugging\n");
  printf("  n <k>    Set value of n to k\n");
  printf("  A        Archive facts\n");
  printf("  B        Retrieve facts\n");
  printf("  C        Archive network\n");
  printf("  D        Retrieve network\n");
  printf("  e        Erase fact from memory\n");
  printf("  c        Clear all facts from memory\n");
  printf("  f        Report on number of facts in memory\n");
  printf("  l        look at a fact in memory\n");
  printf("  L <tag>  look at all facts with given tag\n");
  printf("  z        toggle resizing\n");
  printf("  x        Exit from simulation\n");
  printf("\n");
}

erase_fact(which_fact)		       /* erase a particular fact from
				        * memory */
int which_fact;
{
  int i, fact_index;

  for (fact_index = which_fact; fact_index < next_fact - 1; fact_index++) {
    for (i = 0; i < num_words; i++) {
      facts[fact_index].pattern[i] = facts[fact_index + 1].pattern[i];
    }
    for (i = 0; i < TAG_SIZE; ++i)
      facts[fact_index].tag[i] = facts[fact_index + 1].tag[i];
  }
  --next_fact;
}



get_parameters()
/*	Get all parameters for network.		*/

{
  int i;
  char temp, c;
  char line[MAX_STRLEN];
  char param_filename[MAX_STRLEN];
  FILE *pfd;

  printf("Do you want enter parameters manually (y/n)?\n");
  gets(line);
  sscanf(line, "%c", &temp);
  if ((temp == 'N') || (temp == 'n'))
    show_prompts = FALSE;
  else
    show_prompts = TRUE;
  if (!show_prompts) {
    printf("Parameter filename?\n");
    gets(line);
    sscanf(line, "%s", param_filename);
    pfd = fopen(param_filename, "r");
    if (pfd == NULL) {
      printf("Error opening %s!\n", param_filename);
      exit(1);
    }
    printf("Reading parameters from %s...\n", param_filename);
  }
  else
    pfd = stdin;
  if (show_prompts)
    printf("Random number seed?\n");
  fgets(line, MAX_STRLEN, pfd);
  sscanf(line, "%d", &seed);
  srand(seed);
  if (show_prompts)
    printf("Number of layers?\n");
  fgets(line, MAX_STRLEN, pfd);
  sscanf(line, "%d", &nlayer);
  if (nlayer > MAX_LAYER) {
    printf("ERROR: Too many layers requested, %d maximum\n", MAX_LAYER);
    exit(1);
  }
  for (i = 0; i < nlayer; i++) {
    if (show_prompts)
      printf("Number of neurons in layer %d?\n", i);
    fgets(line, MAX_STRLEN, pfd);
    sscanf(line, "%d", &nl[i]);
    if (nl[i] > MAX_NEURONS_PER_LAYER) {
      printf("ERROR: Too many neurons/layer requested, %d maximum\n", MAX_NEURONS_PER_LAYER);
      exit(1);
    }
  }
  if (show_prompts)
    printf("Learning rate?\n");
  fgets(line, MAX_STRLEN, pfd);
  sscanf(line, "%f", &learning_rate);
  if (show_prompts)
    printf("Momentum rate?\n");
  fgets(line, MAX_STRLEN, pfd);
  sscanf(line, "%f", &momentum_rate);

  if (pfd != stdin)
    fclose(pfd);

  initialize_network();
}				       /* get_parameters */



float random_uniform()
{
  return ((float) rand() / MAX_LONG);
}

initialize_network()
{
  int i, j, k;

  for (i = 1; i < nlayer; i++) {
    for (j = 0; j < nl[i]; j++) {
      /* k=nl[i-1] is for connection to hardwired "ON" unit */
      for (k = 0; k <= nl[i - 1]; k++) {
	net[i][j].connection[k] = -0.2 + (0.4 * random_uniform());
      }
      net[i][j].output = 0;
    }
  }
}				       /* initialize_network */




get_fact_from_net(factp)
struct a_fact *factp;
{
  int j, max_j;
  float max, max2;

  max = -1;			       /* ouputs should be from 0 to 1 */
  max2 = 0;
  max_j = 0;
  for (j = 0; j < nl[nlayer - 1]; j++)
    if (net[nlayer - 1][j].output > max) {
      max2 = max;
      max = net[nlayer - 1][j].output;
      max_j = j;
    }
    else if (net[nlayer - 1][j].output > max2)
      max2 = net[nlayer - 1][j].output;
  factp->tag[0] = 'A' + max_j;
  printf("Confidence of answer: %4.2f, %4.2f above next answer.\n", max, (max - max2));
}				       /* get_fact_from_net */



back_propagate(factp)
struct a_fact *factp;
{
  register int i, j, k;
  register float temp, temp2;
  register a_site *net_i_j;

/* set error of last layer */

  for (j = 0; j < nl[nlayer - 1]; j++) {
    temp = net[nlayer - 1][j].output;
    temp2 = ((factp->tag[0] - 'A') == j ? 1.0 : 0.0);
    net[nlayer - 1][j].output = (temp2 - temp) * temp * (1 - temp);
  }

  for (i = nlayer - 1; i > 0; i--) {

/* compute the change in connectivity strength */

    for (j = 0; j < nl[i]; j++) {
      net_i_j = &net[i][j];
      for (k = 0; k < nl[i - 1]; k++) {
	net_i_j->delta[k] = learning_rate * net_i_j->output * net[i - 1][k].output + momentum_rate * net_i_j->delta[k];
      }
      net_i_j->delta[nl[i - 1]] = learning_rate * net_i_j->output + momentum_rate * net_i_j->delta[nl[i - 1]];
    }

/* set error of previous layer */

    for (j = 0; j < nl[i - 1]; j++) {
      temp = 0;
      for (k = 0; k < nl[i]; k++) {
	temp += net[i][k].output * net[i][k].connection[j];
      }
      temp2 = net[i - 1][j].output;
      net[i - 1][j].output = temp2 * (1.0 - temp2) * temp;
    }

/* and update connections */

    for (j = 0; j < nl[i]; j++) {
      net_i_j = &net[i][j];
      for (k = 0; k <= nl[i - 1]; k++) {
	net_i_j->connection[k] += net_i_j->delta[k];
      }
    }

  }
}				       /* back_propagate */
